#include <microsmooth.h>
#include <OneButton.h>
#include <TimedAction.h>
#include <SoftwareSerial.h>

SoftwareSerial lcd(23, 12);

uint16_t *delayHistory     = ms_init(EMA);

// Analog inputs
const int delayInputPin     = A0;
const int thresholdInputPin = A5;
const int audioInputPin     = A1;
const int lightSensorPin    = A4;

// Digital inputs
const int manualTriggerPin  = 5;
const int laserTogglePin    = 7;

// Digital outputs
const int buzzerPin            =  2;
const int laserPin             = 11;
const int LED_audioTriggerPin  =  4;
const int LED_lightTriggerPin  =  6;
const int LED_statusPin        = 13;

const int cameraPin            =  9;
const int flashPin             = 10;

const byte STATUS_IDLE        = 0;
const byte STATUS_ARMED_SOUND = 1;
const byte STATUS_ARMED_LIGHT = 2;

byte status = STATUS_IDLE;

const int TRIGGER_SOUND = 0;
const int TRIGGER_LIGHT = 1;

boolean laserIsOn = false;
boolean laserButtonPreviousState = false;

long triggerMillis[2]; // timestamp of when sensors were triggered

OneButton laserToggleButton(laserTogglePin, false);
OneButton modeButton(manualTriggerPin, false);

int lightThreshold = 100;
int shutterDelay;

boolean statusLEDisOn = false;

TimedAction soundStatusBlinkAction = TimedAction(40, blinkStatusLED);
TimedAction lightStatusBlinkAction = TimedAction(2000, blinkStatusLED);


void setup() {
  
  lcd.begin(9600);
  delay(500);
  
  triggerMillis[0] = 0;
  triggerMillis[1] = 0;
  
  Serial.begin(9600);

  pinMode(laserTogglePin, INPUT);
  
  pinMode(laserPin, OUTPUT);
  digitalWrite(laserPin, LOW);  

  pinMode(cameraPin, OUTPUT);
  digitalWrite(cameraPin, LOW);
  
  pinMode(flashPin, OUTPUT);
  digitalWrite(flashPin, LOW);

  pinMode(LED_statusPin, OUTPUT);
  digitalWrite(LED_statusPin, HIGH);

  pinMode(LED_audioTriggerPin, OUTPUT);
  digitalWrite(LED_audioTriggerPin, HIGH);

  pinMode(manualTriggerPin, INPUT);
  pinMode(LED_lightTriggerPin, OUTPUT);

  modeButton.setClickTicks(300);
  modeButton.attachClick(singleClickHandler);
  modeButton.attachDoubleClick(doubleClickHandler);

  laserToggleButton.setClickTicks(300);
  laserToggleButton.attachClick(toggleLaser);

  flashLEDs();
  flashLEDs();
  flashLEDs();
  flashLEDs();
  flashLEDs();
}

void loop() {
  modeButton.tick();
  
  switch (status)
  {
    case(STATUS_IDLE):

      detectLasterToggle();
      adjustDelay();
      detectTrigger(TRIGGER_LIGHT); 
      detectTrigger(TRIGGER_SOUND); 
      break;

    case(STATUS_ARMED_SOUND):
      
      soundStatusBlinkAction.check();
      
      if (!shutterIsOpen())
      {
        openShutter();
      } 
  
      if (detectTrigger(TRIGGER_SOUND)) {
        delay(shutterDelay);
        triggerPin(flashPin);
        closeShutter();
        status = STATUS_IDLE;
      }
      
      break;
      
    case(STATUS_ARMED_LIGHT):
    
      turnLaserOn();
      lightStatusBlinkAction.check();
      
      if (!shutterIsOpen())
      {
        openShutter();
      } 
  
      if (detectTrigger(TRIGGER_LIGHT)) {
        delay(shutterDelay);
        turnLaserOff();
        triggerPin(flashPin);
        closeShutter();
        status = STATUS_IDLE;
      }
       
      break;
  }

}

void detectLasterToggle()
{
  boolean currentLaserButtonState = digitalRead(laserTogglePin);

  if (currentLaserButtonState != laserButtonPreviousState) 
  {
    // debounce
    delay(10);
  }
  
  if (currentLaserButtonState == HIGH && laserButtonPreviousState == LOW)
  {
    laserIsOn = !laserIsOn;
  }

  laserButtonPreviousState = currentLaserButtonState;
  digitalWrite(laserPin, laserIsOn);
}

void adjustDelay()
{
  shutterDelay = ema_filter(analogRead(delayInputPin), delayHistory);
  shutterDelay = map(shutterDelay, 0, 1023, 0, 100);
  char delayString[8];
  sprintf(delayString,"%8d", shutterDelay);
  
  lcd.write(254); // move cursor to beginning of second line
  lcd.write(192);
  lcd.write("Delay ");
  lcd.write(delayString);
  lcd.write("ms");
}

boolean shutterIsOpen()
{
  return digitalRead(cameraPin) == HIGH;
}

void openShutter()
{
  digitalWrite(cameraPin, HIGH);
  delay(10);  
}

void closeShutter()
{
  digitalWrite(cameraPin, LOW);
}

boolean detectTrigger(int type)
{
  boolean hasTriggered = false;
  int sensor;
  int threshold;
  int ledPin;
  
  if (type == 0) 
  {
    // audio
    sensor = analogRead(audioInputPin);
    threshold = analogRead(thresholdInputPin);
    ledPin = LED_audioTriggerPin;

    char tempstring[4];
    sprintf(tempstring,"%4d",threshold);

    lcd.write(254); // move cursor to beginning of first line
    lcd.write(128);
    lcd.write("Snd thresh. ");
    lcd.write(tempstring);
  
  } 
  else if (type == 1) 
  {
    // light
    sensor =  analogRead(lightSensorPin);
    threshold = lightThreshold;
    ledPin = LED_lightTriggerPin;
  }
  
  if (sensor > threshold) 
  {
    digitalWrite(ledPin, LOW);
    triggerMillis[type] = millis();
    hasTriggered = true;
  }
  
  if (digitalRead(ledPin) == LOW && millis() - triggerMillis[type] >= 100)
  {
    digitalWrite(ledPin, HIGH);
  }
  
  return hasTriggered;
}

void triggerPin(int pin)
{
  digitalWrite(pin, HIGH);
  delay(10);
  digitalWrite(pin, LOW);
}

void flashLEDs()
{
  digitalWrite(LED_lightTriggerPin, LOW);
  digitalWrite(LED_audioTriggerPin, LOW);
  digitalWrite(LED_statusPin, LOW);
  delay(50);
  digitalWrite(LED_lightTriggerPin, HIGH);
  digitalWrite(LED_audioTriggerPin, HIGH);
  digitalWrite(LED_statusPin, HIGH);
  delay(100);
}

void singleClickHandler() {
  if (detectTrigger(TRIGGER_LIGHT)) {
    // Can't arm light mode if the light sensor has already tripped
    toneError();
    return;
  }
  
  status = STATUS_ARMED_LIGHT;
  toneRising();
  delay(100);  
}

void doubleClickHandler() {
  status = STATUS_ARMED_SOUND;
  toneDescending();
  delay(200);  

}

void toneError()
{
  for (int i = 1; i <= 3; i++) {
    tone(buzzerPin, 2000, 100);
    delay(200);
  }
}

void toneRising()
{
  for (int i = 100; i <= 10000; i += 5) {
    tone(buzzerPin, i, 5);
  }
}

void toneDescending()
{
  for (int i = 10000; i >= 100; i -= 5) {
    tone(buzzerPin, i, 5);
  }
}

void blinkStatusLED()
{
  statusLEDisOn = !statusLEDisOn;
  digitalWrite(LED_statusPin, statusLEDisOn);
}

void toggleLaser()
{
  laserIsOn = !laserIsOn;
  digitalWrite(laserPin, laserIsOn);
  Serial.println(laserIsOn);

}

void turnLaserOn()
{
  digitalWrite(laserPin, HIGH);
}

void turnLaserOff()
{
  digitalWrite(laserPin, LOW);
}
